#include <stdio.h>
#include <libguile.h>
#include <guile/gh.h>

typedef long int (*ghp)();

extern void _XEditResCheckMessages();

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

#include <Xm/Xm.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/Form.h>
#include <Xm/ToggleB.h>
#include <Xm/DrawingA.h>
#include <Xm/Frame.h>
#include <Xm/AtomMgr.h>
#include <Xm/Protocols.h>
#include <Xm/CascadeB.h>
#include <Xm/DialogS.h>
#include <Xm/TextF.h>
#include <Xm/Text.h>
#include <Xm/ScrollBar.h>
#include <Xm/FileSBP.h>
#include <Xm/List.h>


#include "motif.h"

static int debug=0;


char *gh_scm2string (SCM str)
{
  char *cstr;
  int len;
  /* gh_scm2str (str, &cstr, &len); */

  cstr = gh_scm2newstr (str, &len);

  /* why do I care about len?  is this thing not being
     NULL terminated? */

  /* FIX THIS */

  /*return strdup (cstr);*/
  return cstr;
}


KeySym Key2Keysym (XKeyEvent *keyevent)
{
    XComposeStatus composestatus;
    KeySym keysym;
    char string[10];
    
    XLookupString(keyevent, string, 9, &keysym, &composestatus);
    return(keysym);
}


SCM c_string_to_scm_string (SCM scm_c_string)
{
  char *c_string = (char *) gh_scm2ulong (scm_c_string);
  return gh_str02scm (c_string);
}


SCM scm_string_to_c_string (SCM scm_string)
{
  char *c_string = gh_scm2string (scm_string);
  /* need strdup? */
  return gh_ulong2scm ((unsigned long) strdup (c_string));
}


SCM gh_XtInitialize (SCM shell_name, SCM app_class)
{
  Widget top_level;
  int argc=0;
  char *c_shell_name;
  char *c_app_class;


  if (debug) printf ("gh_XtInitialize\n");
  
  gh_defer_ints();

  c_shell_name = gh_scm2string (shell_name);
  c_app_class = gh_scm2string (app_class);

  top_level = XtInitialize (c_shell_name, c_app_class, NULL, 0, &argc, NULL);
  /* free (c_shell_name); free (c_app_class); */

  XtAddEventHandler (top_level, (EventMask)0, True,
		     (XtEventHandler)_XEditResCheckMessages, NULL);


  if (debug) printf (" gh_XtInitialize done\n");

  gh_allow_ints();

  return gh_ulong2scm ((unsigned long) top_level);
  /* didn't you know a Widget was an unsigned long? */
}


SCM gh_GetApplicationContext (SCM top_level)
{
  XtAppContext app_context;
  Display *display = XtDisplay ((Widget) gh_scm2ulong (top_level));
  char *fallbacks[] = {"*background:grey30",
		       "*XmPushButton*background:grey30",
		       NULL};

  gh_defer_ints();
  if (debug) printf (" gh_GetApplicationContext\n");

  app_context = XtDisplayToApplicationContext (display);
  XtAppSetFallbackResources (app_context, fallbacks);
  XtSetLanguageProc (NULL, NULL, NULL);


  if (debug) printf (" gh_GetApplicationContext done\n");
  gh_allow_ints();

  return gh_ulong2scm ((unsigned long) app_context);
  /* an XtAppContext is also an unsigned long! */
}


SCM gh_XtVaAppCreateShell (SCM scm_top_level, SCM scm_name)
{
  Widget top_level = (Widget) gh_scm2ulong (scm_top_level);
  char *name;
  Widget shell;

  gh_defer_ints();
  if (debug) printf (" gh_XtVaAppCreateShell\n");


  name = gh_scm2string (scm_name);

  shell = XtVaAppCreateShell(name, name,
			     applicationShellWidgetClass,
			     XtDisplay (top_level),
			     XmNdeleteResponse, XmDO_NOTHING,
			     XmNallowShellResize, False,
			     /*
			     XmNwidth, 400,
			     XmNheight, 400,
			     */
			     NULL);

  /* free (name); */

  XtAddEventHandler (top_level, (EventMask)0, True,
		     (XtEventHandler)_XEditResCheckMessages, NULL);

  if (debug) printf (" gh_XtVaAppCreateShell done\n");
  gh_allow_ints();

  return gh_ulong2scm ((unsigned long) shell);
}


SCM gh_XtSetKeyboardFocus (SCM scm_subtree, SCM scm_descendant)
{
  Widget subtree = (Widget) gh_scm2ulong (scm_subtree);
  Widget descendant = (Widget) gh_scm2ulong (scm_descendant);

  gh_defer_ints();
  if (debug) printf (" gh_XtSetKeyboardFocus\n");

  XtSetKeyboardFocus (subtree, descendant);

  /*
  XSetInputFocus (XtDisplay (descendant),
		  XtWindow (descendant),
		  RevertToParent,
		  CurrentTime);
		  */


  if (debug) printf (" gh_XtSetKeyboardFocus done\n");
  gh_allow_ints();

  return gh_ulong2scm ((unsigned long) scm_descendant);
}


SCM gh_XtRealizeWidget (SCM scm_widget)
{
  Widget w = (Widget) gh_scm2ulong (scm_widget);

  gh_defer_ints();
  if (debug) printf (" gh_XtRealizeWidget\n");

  XtRealizeWidget (w);

  if (debug) printf (" gh_XtRealizeWidget done\n");
  gh_allow_ints();

  return scm_widget;
}


SCM gh_XmCreateMenuBar (SCM scm_widget)
{
  Widget w = (Widget) gh_scm2ulong (scm_widget);
  Widget rc;

  gh_defer_ints();
  if (debug) printf (" gh_XmCreateMenuBar\n");

  rc = XmCreateMenuBar (w, "menubar", NULL, 0);

  if (debug) printf (" gh_XmCreateMenuBar done\n");
  gh_allow_ints();


  return gh_ulong2scm ((unsigned long) rc);
}



SCM gh_XmCreatePulldownMenu (SCM scm_menu_bar)
{
  Widget menu_bar = (Widget) gh_scm2ulong (scm_menu_bar);
  Widget menu;

  gh_defer_ints();
  if (debug) printf (" gh_XmCreatePulldownMenu\n");

  menu = XmCreatePulldownMenu (menu_bar, "menu", NULL, 0);

  if (debug) printf (" gh_XmCreatePulldownMenu done\n");
  gh_allow_ints();

  return gh_ulong2scm ((unsigned long) menu);
}


void gh_XtCallback (Widget w, XtPointer client_data, XtPointer call_data)
{
  SCM proc_data = (SCM) client_data;
  SCM proc, scm_client_data, scm_call_data, scm_widget;
  SCM args;

  gh_defer_ints();
  if (debug) printf (" gh_XtCallback\n");

  proc = gh_car (proc_data);

  scm_client_data = gh_cadr (proc_data);
  scm_call_data = gh_ulong2scm ((unsigned long) call_data);
  scm_widget = gh_ulong2scm ((unsigned long) w);
  /* hmmm... */
  /* args = gh_list (scm_client_data, scm_call_data, SCM_EOL);*/
  args = gh_list (scm_widget, scm_client_data, scm_call_data, SCM_UNDEFINED);


  scm_apply (proc, args, SCM_EOL);
  /* gh_apply (proc, args); ??? */

  if (debug) printf (" gh_XtCallback done\n");
  gh_allow_ints();

}


SCM gh_CallDataXEvent (SCM scm_call_data)
{
  XtPointer call_data;
  XEvent *e;

  gh_defer_ints();
  if (debug) printf (" gh_CallDataXEvent\n");

  call_data = (XtPointer) gh_scm2ulong (scm_call_data);
  e = ((XmAnyCallbackStruct *) call_data)->event;

  if (debug) printf (" gh_CallDataXEvent done\n");
  gh_allow_ints();

  return gh_ulong2scm ((unsigned long) e);
}


SCM unmangleXEvent (SCM scm_event)
{
  XEvent *e;
  SCM ret = SCM_BOOL_F;

  gh_defer_ints();
  if (debug) printf (" gh_unmangleXEvent\n");

  e = (XEvent *) gh_scm2ulong (scm_event);

  if (e != NULL)
    switch (e->type)
      {
      case KeyPress:
      case KeyRelease:
	{
	  XKeyEvent *xke = (XKeyEvent *) e;
	  KeySym keysym = Key2Keysym (xke);
	  if (debug)
	    printf ("it is a keypress\n");
	  ret = gh_list (gh_ulong2scm (xke->type),
			 gh_long2scm (xke->state),
			 /* gh_char2scm ((char) (keysym)), */
			 gh_ulong2scm (keysym),
			 gh_long2scm (xke->keycode),
			 SCM_UNDEFINED);
	  break;
	}
      case ButtonPress:
	{
	  XButtonEvent *xbe = (XButtonEvent *) e;
	  if (debug)
	    printf ("it is a button press\n");
	  ret = gh_list (gh_ulong2scm (xbe->type),
			 gh_long2scm (xbe->x), gh_long2scm (xbe->y),
			 gh_long2scm (xbe->x_root), gh_long2scm (xbe->y_root),
			 gh_long2scm (xbe->state),
			 gh_ulong2scm (xbe->button),
			 SCM_UNDEFINED);
	  break;
	}
      case Expose:
	{
	  ret = gh_list (gh_ulong2scm (e->xexpose.window),
			 gh_ulong2scm (e->xexpose.x),
			 gh_ulong2scm (e->xexpose.y),
			 gh_ulong2scm (e->xexpose.width),
			 gh_ulong2scm (e->xexpose.height),
			 SCM_UNDEFINED);
	  break;
	}
      case ResizeRequest:
	{
	  ret = gh_list (gh_ulong2scm (e->xresizerequest.window),
			 gh_ulong2scm (e->xresizerequest.width),
			 gh_ulong2scm (e->xresizerequest.height),
			 SCM_UNDEFINED);
	  break;
	}
      default:
	{
	  ret = SCM_BOOL_F;
	  break;
	}
      }
  
  if (debug) printf (" gh_unmangleXEvent done\n");
  gh_allow_ints();

  return ret;
  

      /*
	KeyRelease
	ButtonRelease
	MotionNotify
	EnterNotify
	LeaveNotify
	FocusIn
	FocusOut
	KeymapNotify
	Expose
	GraphicsExpose
	NoExpose
	VisibilityNotify
	CreateNotify
	DestroyNotify
	UnmapNotify
	MapNotify
	MapRequest
	ReparentNotify
	ConfigureNotify
	ConfigureRequest
	GravityNotify
	ResizeRequest
	CirculateNotify
	CirculateRequest
	PropertyNotify
	SelectionClear
	SelectionRequest
	SelectionNotify
	ColormapNotify
	ClientMessage
	MappingNotify
	LASTEvent
	*/
}

SCM gh_XtAddCallback (SCM scm_widget, SCM scm_name, SCM proc, SCM arg)
{
  Widget w;
  char *name;
  SCM proc_arg;

  gh_defer_ints();
  if (debug) printf (" gh_XtAddCallback\n");

  w = (Widget) gh_scm2ulong (scm_widget);
  name = (char *) gh_scm2ulong (scm_name); /* hmmm... */

  proc_arg = gh_cons (proc, arg);

  /* am i tempting g.c. hell? */

  scm_protect_object (proc);
  scm_protect_object (arg);
  scm_protect_object (proc_arg);
  /* gh_gc_mark??? */

  if (debug)
    printf ("proc:%ld, arg:%ld, proc_arg:%ld\n", proc, arg, proc_arg);


  XtAddCallback (w, name, gh_XtCallback, (XtPointer) proc_arg);

  if (debug) printf (" gh_XtAddCallback done\n");
  gh_allow_ints();

  return scm_widget;
}


SCM gh_XmStringCreateLocalized (SCM scm_str)
{
  XmString xmstr;
  char *str;

  gh_defer_ints();
  if (debug) printf (" gh_XmStringCreateLocalized\n");

  str = gh_scm2string (scm_str);
  xmstr = XmStringCreateLocalized (str);

  if (debug) printf (" gh_XmStringCreateLocalized done\n");
  gh_allow_ints();

  return gh_ulong2scm ((unsigned long) xmstr);
}


SCM gh_XmStringFree (SCM scm_xmstr)
{
  XmString xmstr;

  gh_defer_ints();
  if (debug) printf (" gh_XmStringFree\n");

  xmstr = (XmString) gh_scm2ulong (scm_xmstr);
  XmStringFree (xmstr);

  if (debug) printf (" gh_XmStringFree done\n");
  gh_allow_ints();

  return SCM_BOOL_T;
}


SCM gh_XtManageChild (SCM scm_widget)
{
  Widget w = (Widget) gh_scm2ulong (scm_widget);

  gh_defer_ints();
  if (debug) printf (" gh_XtManageChild\n");

  XtManageChild (w);

  if (debug) printf (" gh_XtManageChild done\n");
  gh_allow_ints();

  return scm_widget;
}


SCM gh_XtCreateWidget (SCM scm_name,
		       SCM scm_type,
		       SCM scm_parent,
		       SCM scm_args)
{
  char *name = gh_scm2string (scm_name);
  WidgetClass type = (WidgetClass) gh_scm2ulong (scm_type);
  Widget parent = (Widget) gh_scm2ulong (scm_parent);
  Widget w;
  Arg *wargs = NULL;
  int nargs = 0;

  gh_defer_ints();
  if (debug) printf (" gh_XtCreateWidget\n");

  /* w = XtVaCreateWidget (name, type, parent, NULL); */

  if (gh_list_p (scm_args))
    {
      wargs = (Arg *) XtMalloc (sizeof (Arg) * gh_list_length (scm_args));

      while (gh_list_length (scm_args) > 0)
	{
	  SCM first_pair = gh_car (scm_args);
	  String name = (char *) gh_scm2ulong (gh_car (first_pair));
	  XtArgVal value = (XtArgVal) gh_scm2ulong (gh_cadr (first_pair));

	  XtSetArg (wargs[ nargs ], name, value);
	  nargs++;
	  scm_args = gh_cdr (scm_args);
	}
    }

  w = XtCreateWidget (name, type, parent, wargs, nargs);

  XtFree ((char *) wargs);

  if (debug) printf (" gh_XtCreateWidget done\n");
  gh_allow_ints();

  return gh_ulong2scm ((unsigned long) w);
}


SCM gh_XmCreateScrolledList (SCM scm_parent,
			     SCM scm_name,
			     SCM scm_args)
{
  char *name = gh_scm2string (scm_name);
  Widget parent = (Widget) gh_scm2ulong (scm_parent);
  Widget w;
  Arg *wargs = NULL;
  int nargs = 0;

  gh_defer_ints();
  if (debug) printf (" gh_XmCreateScrolledList\n");

  /* w = XtVaCreateWidget (name, type, parent, NULL); */

  if (gh_list_p (scm_args))
    {
      wargs = (Arg *) XtMalloc (sizeof (Arg) * gh_list_length (scm_args));

      while (gh_list_length (scm_args) > 0)
	{
	  SCM first_pair = gh_car (scm_args);
	  String name = (char *) gh_scm2ulong (gh_car (first_pair));
	  XtArgVal value = (XtArgVal) gh_scm2ulong (gh_cadr (first_pair));

	  XtSetArg (wargs[ nargs ], name, value);
	  nargs++;
	  scm_args = gh_cdr (scm_args);
	}
    }

  w = XmCreateScrolledList (parent, name, wargs, nargs);

  XtFree ((char *) wargs);

  if (debug) printf (" gh_XmCreateScrolledList done\n");
  gh_allow_ints();

  return gh_ulong2scm ((unsigned long) w);
}



SCM gh_XmListDeleteAllItems (SCM scm_w)
{
  Widget w = (Widget) gh_scm2ulong (scm_w);

  gh_defer_ints();
  if (debug) printf (" gh_XmListDeleteAllItems\n");

  XmListDeleteAllItems (w);

  if (debug) printf (" gh_XmListDeleteAllItems done\n");
  gh_allow_ints();

  return SCM_BOOL_T;
}


SCM gh_XmListAddItem (SCM scm_w, SCM scm_item, SCM scm_pos)
{
  Widget w = (Widget) gh_scm2ulong (scm_w);
  XmString item = (XmString) gh_scm2ulong (scm_item);
  int pos = gh_scm2int (scm_pos);

  gh_defer_ints();
  if (debug) printf (" gh_XmListAddItem\n");

  XmListAddItem (w, item, pos);

  if (debug) printf (" gh_XmListAddItem done\n");
  gh_allow_ints();

  return SCM_BOOL_T;
}



SCM gh_XtSetValue (SCM scm_widget, SCM scm_name, SCM scm_value)
{
  Widget w;
  unsigned long name, value;

  gh_defer_ints();
  if (debug) printf (" gh_XtSetValue\n");

  w = (Widget) gh_scm2ulong (scm_widget);
  name = gh_scm2ulong (scm_name);
  value = gh_scm2ulong (scm_value);

  XtVaSetValues (w, name, value, NULL, NULL);

  if (debug) printf (" gh_XtSetValue done\n");
  gh_allow_ints();

  return scm_widget;
}


SCM gh_XtGetValue (SCM scm_widget, SCM scm_name)
{
  Widget w;
  unsigned long name, value;

  gh_defer_ints();
  if (debug) printf (" gh_XtGetValue\n");

  w = (Widget) gh_scm2ulong (scm_widget);
  name = gh_scm2ulong (scm_name);

  XtVaGetValues (w, name, &value, NULL, NULL);

  if (debug) printf (" gh_XtGetValue done\n");
  gh_allow_ints();

  return gh_ulong2scm ((unsigned long) value);
}


SCM gh_XtDisplay (SCM scm_widget)
{
  Widget w;
  Display *display;

  gh_defer_ints();
  if (debug) printf (" gh_XtDisplay\n");

  w = (Widget) gh_scm2ulong (scm_widget);

  display = XtDisplay (w);

  if (debug) printf (" gh_XtDisplay done\n");
  gh_allow_ints();

  return gh_ulong2scm ((unsigned long) display);
}


SCM gh_XtWindow (SCM scm_widget)
{
  Widget w;
  Window window;

  gh_defer_ints();
  if (debug) printf (" gh_XtWindow\n");

  w = (Widget) gh_scm2ulong (scm_widget);
  window = XtWindow (w);

  if (debug) printf (" gh_XtWindow done\n");
  gh_allow_ints();

  return gh_ulong2scm ((unsigned long) window);
}


SCM gh_XCreateGC (SCM scm_display, SCM scm_d)
{
  Display *display;
  Drawable d;
  GC gc;

  gh_defer_ints();
  if (debug) printf (" gh_XCreateGC\n");

  display = (Display *) gh_scm2ulong (scm_display);
  d = (Drawable) gh_scm2ulong (scm_d);
  gc = XCreateGC (display, d, 0, NULL);

  if (debug) printf (" gh_XCreateGC done\n");
  gh_allow_ints();

  return gh_ulong2scm ((unsigned long) gc);
}


SCM gh_XFreeGC (SCM scm_display, SCM scm_gc)
{
  Display *display;
  GC gc;

  gh_defer_ints();
  if (debug) printf (" gh_XFreeGC\n");

  display = (Display *) gh_scm2ulong (scm_display);
  gc = (GC) gh_scm2ulong (scm_gc);
  XFreeGC (display, gc);

  if (debug) printf (" gh_XFreeGC done\n");
  gh_allow_ints();

  return SCM_BOOL_T;
}


SCM gh_XDrawRectangle (SCM scm_display,
		       SCM scm_d,
		       SCM scm_gc,
		       SCM scm_x,
		       SCM scm_y,
		       SCM scm_width,
		       SCM scm_height)
{
  Display *display;
  Drawable d;
  GC gc;
  int x, y;
  unsigned int width, height;

  gh_defer_ints();
  if (debug) printf (" gh_XDrawRectangle\n");

  display = (Display *) gh_scm2ulong (scm_display);
  d = (Drawable) gh_scm2ulong (scm_d);
  gc = (GC) gh_scm2ulong (scm_gc);
  x = (int) gh_scm2ulong (scm_x);
  y = (int) gh_scm2ulong (scm_y);
  width = (int) gh_scm2ulong (scm_width);
  height = (int) gh_scm2ulong (scm_height);
  XDrawRectangle (display, d, gc, x, y, width, height);

  if (debug) printf (" gh_XDrawRectangle done\n");
  gh_allow_ints();

  return SCM_BOOL_T;
}


SCM gh_XDrawLine (SCM scm_display,
		  SCM scm_d,
		  SCM scm_gc,
		  SCM scm_x1,
		  SCM scm_y1,
		  SCM scm_x2,
		  SCM scm_y2)
{
  Display *display;
  Drawable d;
  GC gc;
  int x1, y1, x2, y2;

  gh_defer_ints();
  if (debug) printf (" gh_XDrawLine\n");

  display = (Display *) gh_scm2ulong (scm_display);
  d = (Drawable) gh_scm2ulong (scm_d);
  gc = (GC) gh_scm2ulong (scm_gc);
  x1 = (int) gh_scm2ulong (scm_x1);
  y1 = (int) gh_scm2ulong (scm_y1);
  x2 = (int) gh_scm2ulong (scm_x2);
  y2 = (int) gh_scm2ulong (scm_y2);
  XDrawLine (display, d, gc, x1, y1, x2, y2);

  if (debug) printf (" gh_XDrawLine done\n");
  gh_allow_ints();

  return SCM_BOOL_T;
}


SCM gh_XFillRectangle (SCM scm_display,
		       SCM scm_d,
		       SCM scm_gc,
		       SCM scm_x,
		       SCM scm_y,
		       SCM scm_width,
		       SCM scm_height)
{
  Display *display;
  Drawable d;
  GC gc;
  int x, y;
  unsigned int width, height;

  gh_defer_ints();
  if (debug) printf (" gh_XFillRectangle\n");

  display = (Display *) gh_scm2ulong (scm_display);
  d = (Drawable) gh_scm2ulong (scm_d);
  gc = (GC) gh_scm2ulong (scm_gc);
  x = (int) gh_scm2ulong (scm_x);
  y = (int) gh_scm2ulong (scm_y);
  width = (int) gh_scm2ulong (scm_width);
  height = (int) gh_scm2ulong (scm_height);
  XFillRectangle (display, d, gc, x, y, width, height);

  if (debug) printf (" gh_XFillRectangle done\n");
  gh_allow_ints();

  return SCM_BOOL_T;
}


SCM gh_XLoadQueryFont (SCM scm_display, SCM scm_font_name)
{
  Display *display;
  char *font_name;
  XFontStruct *xfs;

  gh_defer_ints();
  if (debug) printf (" gh_XLoadQueryFont\n");

  display = (Display *) gh_scm2ulong (scm_display);
  font_name = gh_scm2string (scm_font_name);

  xfs = XLoadQueryFont (display, font_name);

  if (debug) printf (" gh_XLoadQueryFont done\n");
  gh_allow_ints();

  return gh_ulong2scm ((unsigned long) xfs);
}



SCM gh_XTextWidth (SCM scm_xfs, SCM scm_str)
{
  XFontStruct *xfs;
  char *str;
  unsigned long width;

  gh_defer_ints();
  if (debug) printf (" gh_XTextWidth\n");

  xfs = (XFontStruct *) gh_scm2ulong (scm_xfs);
  str = gh_scm2string (scm_str);

  width = XTextWidth (xfs, str, strlen(str));

  if (debug) printf (" gh_XTextWidth done\n");
  gh_allow_ints();

  return gh_ulong2scm (width);
}


SCM gh_XFontStruct_ascent (SCM scm_xfs)
{
  XFontStruct *xfs;

  if (debug) printf (" gh_XFontStruct_ascent\n");

  xfs = (XFontStruct *) gh_scm2ulong (scm_xfs);

  if (debug) printf (" gh_XFontStruct_ascent done\n");

  return gh_ulong2scm ((unsigned long) xfs->ascent);
}


SCM gh_XFontStruct_descent (SCM scm_xfs)
{
  XFontStruct *xfs;

  if (debug) printf (" gh_XFontStruct_descent\n");

  xfs = (XFontStruct *) gh_scm2ulong (scm_xfs);

  if (debug) printf (" gh_XFontStruct_descent done\n");

  return gh_ulong2scm ((unsigned long) xfs->descent);
}


SCM gh_XFontStruct_fid (SCM scm_xfs)
{
  XFontStruct *xfs;

  if (debug) printf (" gh_XFontStruct_fid\n");

  xfs = (XFontStruct *) gh_scm2ulong (scm_xfs);

  if (debug) printf (" gh_XFontStruct_fid done\n");

  return gh_ulong2scm ((unsigned long) xfs->fid);
}


SCM gh_XDrawString (SCM scm_display,
		    SCM scm_d,
		    SCM scm_gc,
		    SCM scm_x,
		    SCM scm_y,
		    SCM scm_str)
{
  Display *display;
  Drawable d;
  GC gc;
  int x, y;
  char *str;

  gh_defer_ints();
  if (debug) printf (" gh_XDrawString\n");

  display = (Display *) gh_scm2ulong (scm_display);
  d = (Drawable) gh_scm2ulong (scm_d);
  gc = (GC) gh_scm2ulong (scm_gc);
  x = (int) gh_scm2ulong (scm_x);
  y = (int) gh_scm2ulong (scm_y);
  str = gh_scm2string (scm_str);

  XDrawString (display, d, gc, x, y, str, strlen (str));

  if (debug) printf (" gh_XDrawString done\n");
  gh_allow_ints();

  return SCM_BOOL_T;
}


SCM gh_XSetFont (SCM scm_display, SCM scm_gc, SCM scm_font)
{
  Display *display;
  GC gc;
  Font font;

  gh_defer_ints();
  if (debug) printf (" gh_XSetFont\n");

  display = (Display *) gh_scm2ulong (scm_display);
  gc = (GC) gh_scm2ulong (scm_gc);
  font = (Font) gh_scm2ulong (scm_font);

  XSetFont (display, gc, font);

  if (debug) printf (" gh_XSetFont done\n");
  gh_allow_ints();

  return SCM_BOOL_T;
}


SCM gh_DefaultColormap (SCM scm_display)
{
  Display *display;
  Colormap cmap;

  gh_defer_ints();
  if (debug) printf (" gh_DefaultColormap\n");

  display = (Display *) gh_scm2ulong (scm_display);
  cmap = DefaultColormap (display, DefaultScreen(display));

  if (debug) printf (" gh_DefaultColormap done\n");
  gh_allow_ints();

  return gh_ulong2scm ((unsigned long) cmap);
}


SCM gh_new_XColor ()
{
  XColor *xcolorp;

  gh_defer_ints();
  if (debug) printf (" gh_new_XColor\n");

  xcolorp = (XColor *) malloc (sizeof (XColor));

  if (debug) printf (" gh_new_XColor done\n");
  gh_allow_ints();

  return gh_ulong2scm ((unsigned long) xcolorp);
}


SCM gh_XParseColor (SCM scm_display, SCM scm_cmap,
		    SCM scm_color, SCM scm_xcolorp)
{
  Display *display;
  Colormap cmap;
  char *color;
  XColor *xcolorp;
  int result;

  gh_defer_ints();
  if (debug) printf (" gh_XParseColor\n");

  display = (Display *) gh_scm2ulong (scm_display);
  cmap = (Colormap) gh_scm2ulong (scm_cmap);
  color = gh_scm2string (scm_color);
  xcolorp = (XColor *) gh_scm2ulong (scm_xcolorp);

  result = XParseColor (display, cmap, color, xcolorp);

  if (debug) printf (" gh_XParseColor done\n");
  gh_allow_ints();

  if (result == 0)
    return SCM_BOOL_F;

  return SCM_BOOL_T;
}


SCM gh_XAllocColor (SCM scm_display, SCM scm_cmap, SCM scm_xcolorp)
{
  Display *display;
  Colormap cmap;
  XColor *xcolorp;
  int result;

  gh_defer_ints();
  if (debug) printf (" gh_XAllocColor\n");

  display = (Display *) gh_scm2ulong (scm_display);
  cmap = (Colormap) gh_scm2ulong (scm_cmap);
  xcolorp = (XColor *) gh_scm2ulong (scm_xcolorp);

  result = XAllocColor (display, cmap, xcolorp);

  if (debug) printf (" gh_XAllocColor done\n");
  gh_allow_ints();

  if (result == 0)
    return SCM_BOOL_F;

  return SCM_BOOL_T;
}


SCM gh_XColor_pixel (SCM scm_xcolor)
{
  XColor *xcolorp = (XColor *) gh_scm2ulong (scm_xcolor);
  return gh_ulong2scm ((unsigned long) xcolorp->pixel);
}


SCM gh_XSetForeGround (SCM scm_display, SCM scm_gc, SCM scm_xcolorp)
{
  Display *display;
  GC gc;
  XColor *xcolorp;

  gh_defer_ints();
  if (debug) printf (" gh_XSetForeGround\n");

  display = (Display *) gh_scm2ulong (scm_display);
  gc = (GC) gh_scm2ulong (scm_gc);
  xcolorp = (XColor *) gh_scm2ulong (scm_xcolorp);

  XSetForeground (display, gc, xcolorp->pixel);

  if (debug) printf (" gh_XSetForeGround done\n");
  gh_allow_ints();

  return SCM_BOOL_T;
}


SCM gh_XCopyArea (SCM scm_display,
		  SCM scm_src, SCM scm_dest,
		  SCM scm_gc,
		  SCM scm_src_x, SCM scm_src_y,
		  SCM scm_width, SCM scm_height,
		  SCM scm_dest_x, SCM scm_dest_y)
{
  Display *display;
  Drawable src, dest;
  GC gc;
  int src_x, src_y;
  unsigned int width, height;
  int dest_x, dest_y;

  gh_defer_ints();
  if (debug) printf (" gh_XCopyArea\n");

  display = (Display *) gh_scm2ulong (scm_display);
  src = (Drawable) gh_scm2ulong (scm_src);
  dest = (Drawable) gh_scm2ulong (scm_dest);
  gc = (GC) gh_scm2ulong (scm_gc);
  src_x = (int) gh_scm2ulong (scm_src_x);
  src_y = (int) gh_scm2ulong (scm_src_y);
  width = (int) gh_scm2ulong (scm_width);
  height = (int) gh_scm2ulong (scm_height);
  dest_x = (int) gh_scm2ulong (scm_dest_x);
  dest_y = (int) gh_scm2ulong (scm_dest_y);

  XCopyArea (display, src, dest, gc,
	     src_x, src_y, width, height, dest_x, dest_y);

  if (debug) printf (" gh_XCopyArea done\n");
  gh_allow_ints();

  return SCM_BOOL_T;
}



SCM gh_XtPopup (SCM scm_popup_shell, SCM scm_grab_kind)
{
  Widget popup_shell;
  XtGrabKind grab_kind;

  gh_defer_ints();
  if (debug) printf (" gh_XtPopup\n");

  popup_shell = (Widget) gh_scm2ulong (scm_popup_shell);
  grab_kind = (XtGrabKind) gh_scm2ulong (scm_grab_kind);

  XtPopup (popup_shell, grab_kind);

  if (debug) printf (" gh_XtPopup done\n");
  gh_allow_ints();

  return SCM_BOOL_T;
}


SCM gh_XtPopdown (SCM scm_popup_shell)
{
  Widget popup_shell;

  gh_defer_ints();
  if (debug) printf (" gh_XtPopdown\n");

  popup_shell = (Widget) gh_scm2ulong (scm_popup_shell);

  XtPopdown (popup_shell);

  if (debug) printf (" gh_XtPopdown done\n");
  gh_allow_ints();

  return SCM_BOOL_T;
}



SCM gh_XtRemoveAllCallbacks (SCM scm_widget, SCM scm_name)
{
  Widget w;
  char *name;

  gh_defer_ints();
  if (debug) printf (" gh_XtRemoveAllCallbacks\n");

  w = (Widget) gh_scm2ulong (scm_widget);
  name = (char *) gh_scm2ulong (scm_name); /* hmmm... */

  XtRemoveAllCallbacks (w, name);

  if (debug) printf (" gh_XtRemoveAllCallbacks done\n");
  gh_allow_ints();

  return SCM_BOOL_T;
}


SCM gh_XmStringGetLtoR (SCM scm_string, SCM scm_tag)
{
  XmString string;
  XmStringCharSet tag;
  SCM ret;
  char *text;

  gh_defer_ints();
  if (debug) printf (" gh_XmStringGetLtoR\n");

  string = (XmString) gh_scm2ulong (scm_string);
  tag = (XmStringCharSet) gh_scm2ulong (scm_tag);

  
  if (XmStringGetLtoR (string, tag, &text))
    {
      ret = gh_str02scm (strdup (text));
    }
  else
    ret = SCM_BOOL_F;

  if (debug) printf (" gh_XmStringGetLtoR done\n");
  gh_allow_ints();

  return ret;
}


SCM gh_XtMainLoop ()
{
  XtMainLoop();

  return SCM_BOOL_T;
}


SCM gh_dereference (SCM scm_p)
{
  unsigned long *p = (unsigned long *) gh_scm2ulong (scm_p);
  return gh_ulong2scm (*p);
}


/**********************************************************************/
/*    motif constants                                                 */
/**********************************************************************/

#define rghu2s(x) return gh_ulong2scm ((unsigned long) x)

SCM gh_XtNwidth () {rghu2s (XtNwidth);}
SCM gh_XtNheight () {rghu2s (XtNheight);}
SCM gh_XmNvalue () {rghu2s (XmNvalue);}
SCM gh_XmNSubMenuId () {rghu2s (XmNsubMenuId);}
SCM gh_XmNlabelString () {rghu2s (XmNlabelString);}
SCM gh_XmNselectedItems () {rghu2s (XmNselectedItems);}

SCM gh_XmNorientation () {rghu2s (XmNorientation);}
SCM gh_XmVERTICAL () {rghu2s (XmVERTICAL);}
SCM gh_XmHORIZONTAL () {rghu2s (XmHORIZONTAL);}

SCM gh_XmNleftAttachment () {rghu2s (XmNleftAttachment);}
SCM gh_XmNtopAttachment () {rghu2s (XmNtopAttachment);}
SCM gh_XmNrightAttachment () {rghu2s (XmNrightAttachment);}
SCM gh_XmNbottomAttachment () {rghu2s (XmNbottomAttachment);}

SCM gh_XmNleftWidget () {rghu2s (XmNleftWidget);}
SCM gh_XmNtopWidget () {rghu2s (XmNtopWidget);}
SCM gh_XmNrightWidget () {rghu2s (XmNrightWidget);}
SCM gh_XmNbottomWidget () {rghu2s (XmNbottomWidget);}

SCM gh_XmNcolumns () {rghu2s (XmNcolumns);}
SCM gh_XmNvisibleItemCount () {rghu2s (XmNvisibleItemCount);}

SCM gh_XmATTACH_FORM () {rghu2s (XmATTACH_FORM);}
SCM gh_XmATTACH_NONE () {rghu2s (XmATTACH_NONE);}
SCM gh_XmATTACH_WIDGET () {rghu2s (XmATTACH_WIDGET);}

SCM gh_XmNdialogType () {rghu2s (XmNdialogType);}
SCM gh_XmDIALOG_FILE_SELECTION ()
                            {rghu2s (XmDIALOG_FILE_SELECTION);}
SCM gh_XmNdirSpec () {rghu2s (XmNdirSpec);}
SCM gh_XmFONTLIST_DEFAULT_TAG () {rghu2s (XmFONTLIST_DEFAULT_TAG);}

SCM gh_XtGrabNone () {rghu2s (XtGrabNone);}
SCM gh_XtGrabNonexclusive () {rghu2s (XtGrabNonexclusive);}
SCM gh_XtGrabExclusive () {rghu2s (XtGrabExclusive);}


SCM gh_XmNactivateCallback () {rghu2s (XmNactivateCallback);}
SCM gh_XmNdecrementCallback () {rghu2s (XmNdecrementCallback);}
SCM gh_XmNdragCallback () {rghu2s (XmNdragCallback);}
SCM gh_XmNexposeCallback () {rghu2s (XmNexposeCallback);}
SCM gh_XmNincrementCallback () {rghu2s (XmNincrementCallback);}
SCM gh_XmNinputCallback () {rghu2s (XmNinputCallback);}
SCM gh_XmNpageDecrementCallback () {rghu2s (XmNpageDecrementCallback);}
SCM gh_XmNpageIncrementCallback () {rghu2s (XmNpageIncrementCallback);}
SCM gh_XmNresizeCallback () {rghu2s (XmNresizeCallback);}
SCM gh_XmNvalueChangedCallback () {rghu2s (XmNvalueChangedCallback);}
SCM gh_XmNokCallback () {rghu2s (XmNokCallback);}
SCM gh_XmNcancelCallback () {rghu2s (XmNcancelCallback);}
SCM gh_XmNbrowseSelectionCallback () {rghu2s (XmNbrowseSelectionCallback);}

SCM gh_xmDrawingAreaWidgetClass () {rghu2s (xmDrawingAreaWidgetClass);}
SCM gh_xmFormWidgetClass () {rghu2s (xmFormWidgetClass);}
SCM gh_xmPushButtonWidgetClass () {rghu2s (xmPushButtonWidgetClass);}
SCM gh_xmCascadeButtonWidgetClass () {rghu2s (xmCascadeButtonWidgetClass);}
SCM gh_xmScrollBarWidgetClass () {rghu2s (xmScrollBarWidgetClass);}
SCM gh_xmFileSelectionBoxWidgetClass ()
                                {rghu2s (xmFileSelectionBoxWidgetClass);}
SCM gh_xmTextFieldWidgetClass () {rghu2s (xmTextFieldWidgetClass);}



/**********************************************************************/


void hook_up_guile_motif ()
{
  gh_new_procedure1_0 ("c-string-to-scm-string",
		       (ghp) c_string_to_scm_string);
  gh_new_procedure1_0 ("scm-string-to-c-string",
		       (ghp) scm_string_to_c_string);

  gh_new_procedure2_0 ("xt-initialize",
		       (ghp) gh_XtInitialize);
  gh_new_procedure1_0 ("xt-get-application-context",
		       (ghp) gh_GetApplicationContext);
  gh_new_procedure2_0 ("xt-app-create-shell", (ghp) gh_XtVaAppCreateShell);
  gh_new_procedure2_0 ("xt-set-keyboard-focus", (ghp) gh_XtSetKeyboardFocus);
  gh_new_procedure1_0 ("xt-realize-widget", (ghp) gh_XtRealizeWidget);
  gh_new_procedure1_0 ("xt-create-menu-bar", (ghp) gh_XmCreateMenuBar);
  gh_new_procedure1_0 ("xt-create-pulldown-menu",
		       (ghp) gh_XmCreatePulldownMenu);
  gh_new_procedure1_0 ("xm-string-create-localized",
		       (ghp) gh_XmStringCreateLocalized);
  gh_new_procedure1_0 ("xm-string-free", (ghp) gh_XmStringFree);

  gh_new_procedure3_0 ("xm-create-scrolled-list",
		       (ghp) gh_XmCreateScrolledList);
  gh_new_procedure1_0 ("xm-list-delete-all-items",
		       (ghp) gh_XmListDeleteAllItems);
  gh_new_procedure3_0 ("xm-list-add-item", (ghp) gh_XmListAddItem);

  gh_new_procedure4_0 ("xt-add-callback", (ghp) gh_XtAddCallback);
  gh_new_procedure4_0 ("xt-create-widget", (ghp) gh_XtCreateWidget);
  gh_new_procedure1_0 ("xt-manage-child", (ghp) gh_XtManageChild);
  gh_new_procedure3_0 ("xt-set-value", (ghp) gh_XtSetValue);
  gh_new_procedure2_0 ("xt-get-value", (ghp) gh_XtGetValue);
  gh_new_procedure1_0 ("xt-display", (ghp) gh_XtDisplay);
  gh_new_procedure1_0 ("xt-window", (ghp) gh_XtWindow);
  gh_new_procedure2_0 ("xt-create-gc", (ghp) gh_XCreateGC);
  gh_new_procedure2_0 ("xt-free-gc", (ghp) gh_XFreeGC);
  gh_new_procedure ("xt-draw-rectangle", (ghp) gh_XDrawRectangle, 7, 0, 0);
  gh_new_procedure ("xt-fill-rectangle", (ghp) gh_XFillRectangle, 7, 0, 0);
  gh_new_procedure ("xt-draw-line", (ghp) gh_XDrawLine, 7, 0, 0);
  gh_new_procedure ("xt-copy-area", (ghp) gh_XCopyArea, 10, 0, 0);

  gh_new_procedure2_0 ("xt-load-query-font", (ghp) gh_XLoadQueryFont);
  gh_new_procedure2_0 ("xt-text-width", (ghp) gh_XTextWidth);
  gh_new_procedure1_0 ("xt-font-ascent", (ghp) gh_XFontStruct_ascent);
  gh_new_procedure1_0 ("xt-font-descent", (ghp) gh_XFontStruct_descent);
  gh_new_procedure1_0 ("xt-font-fid", (ghp) gh_XFontStruct_fid);
  gh_new_procedure ("xt-draw-string", (ghp) gh_XDrawString, 6, 0, 0);
  gh_new_procedure3_0 ("xt-set-font", (ghp) gh_XSetFont);
  gh_new_procedure1_0 ("xt-default-colormap", (ghp) gh_DefaultColormap);
  gh_new_procedure0_0 ("xt-new-xcolor", (ghp) gh_new_XColor);
  gh_new_procedure4_0 ("xt-parse-color", (ghp) gh_XParseColor);
  gh_new_procedure3_0 ("xt-alloc-color", (ghp) gh_XAllocColor);
  gh_new_procedure1_0 ("xt-xcolor-pixel", (ghp) gh_XColor_pixel);

  gh_new_procedure3_0 ("xt-set-foreground", (ghp) gh_XSetForeGround);

  gh_new_procedure1_0 ("xt-call-data-event", (ghp) gh_CallDataXEvent);
  gh_new_procedure1_0 ("xt-unmangle-x-event", (ghp) unmangleXEvent);

  gh_new_procedure0_0 ("xt-n-width", (ghp) gh_XtNwidth);
  gh_new_procedure0_0 ("xt-n-height", (ghp) gh_XtNheight);
  gh_new_procedure0_0 ("xm-n-value", (ghp) gh_XmNvalue);
  gh_new_procedure0_0 ("xm-n-sub-menu-id", (ghp) gh_XmNSubMenuId);
  gh_new_procedure0_0 ("xm-n-label-string", (ghp) gh_XmNlabelString);

  gh_new_procedure0_0 ("xm-n-selected-items", (ghp) gh_XmNselectedItems);

  gh_new_procedure0_0 ("xm-n-orientation", (ghp) gh_XmNorientation);
  gh_new_procedure0_0 ("xm-vertical", (ghp) gh_XmVERTICAL);
  gh_new_procedure0_0 ("xm-horizontal", (ghp) gh_XmHORIZONTAL);

  gh_new_procedure0_0 ("xm-n-columns", (ghp) gh_XmNcolumns);
  gh_new_procedure0_0 ("xm-n-visible-item-count",
		       (ghp) gh_XmNvisibleItemCount);


  gh_new_procedure0_0 ("xm-n-left-attachment", (ghp) gh_XmNleftAttachment);
  gh_new_procedure0_0 ("xm-n-top-attachment", (ghp) gh_XmNtopAttachment);
  gh_new_procedure0_0 ("xm-n-right-attachment", (ghp) gh_XmNrightAttachment);
  gh_new_procedure0_0 ("xm-n-bottom-attachment",
		       (ghp) gh_XmNbottomAttachment);

  gh_new_procedure0_0 ("xm-n-left-widget", (ghp) gh_XmNleftWidget);
  gh_new_procedure0_0 ("xm-n-top-widget", (ghp) gh_XmNtopWidget);
  gh_new_procedure0_0 ("xm-n-right-widget", (ghp) gh_XmNrightWidget);
  gh_new_procedure0_0 ("xm-n-bottom-widget", (ghp) gh_XmNbottomWidget);

  gh_new_procedure0_0 ("xm-attach-form", (ghp) gh_XmATTACH_FORM);
  gh_new_procedure0_0 ("xm-attach-none", (ghp) gh_XmATTACH_NONE);
  gh_new_procedure0_0 ("xm-attach-widget", (ghp) gh_XmATTACH_WIDGET);

  gh_new_procedure0_0 ("xm-n-dialog-type", (ghp) gh_XmNdialogType);
  gh_new_procedure0_0 ("xm-dialog-file-selection",
		       (ghp) gh_XmDIALOG_FILE_SELECTION);

  gh_new_procedure0_0 ("xm-n-dir-spec", (ghp) gh_XmNdirSpec);
  gh_new_procedure0_0 ("xm-font-list-default-tag",
		       (ghp) gh_XmFONTLIST_DEFAULT_TAG);

  gh_new_procedure2_0 ("xt-popup", (ghp) gh_XtPopup);
  gh_new_procedure1_0 ("xt-popdown", (ghp) gh_XtPopdown);
  gh_new_procedure2_0 ("xt-remove-all-callbacks",
		       (ghp) gh_XtRemoveAllCallbacks);
  gh_new_procedure2_0 ("xm-string-get-l-to-r", (ghp) gh_XmStringGetLtoR);

  gh_new_procedure0_0 ("xt-grab-none", (ghp) gh_XtGrabNone);
  gh_new_procedure0_0 ("xt-grab-none-exclusive",
		       (ghp) gh_XtGrabNonexclusive);
  gh_new_procedure0_0 ("xt-grab-exclusive", (ghp) gh_XtGrabExclusive);


  gh_new_procedure0_0 ("xm-n-expose-callback", (ghp) gh_XmNexposeCallback);
  gh_new_procedure0_0 ("xm-n-input-callback", (ghp) gh_XmNinputCallback);
  gh_new_procedure0_0 ("xm-n-resize-callback", (ghp) gh_XmNresizeCallback);
  gh_new_procedure0_0 ("xm-n-activate-callback", (ghp) gh_XmNactivateCallback);
  gh_new_procedure0_0 ("xm-n-value-changed-callback",
		       gh_XmNvalueChangedCallback);
  gh_new_procedure0_0 ("xm-n-drag-callback", (ghp) gh_XmNdragCallback);
  gh_new_procedure0_0 ("xm-n-increment-callback",
		       (ghp) gh_XmNincrementCallback);
  gh_new_procedure0_0 ("xm-n-page-increment-callback",
		       gh_XmNpageIncrementCallback);
  gh_new_procedure0_0 ("xm-n-decrement-callback",
		       (ghp) gh_XmNdecrementCallback);
  gh_new_procedure0_0 ("xm-n-page-decrement-callback",
		       gh_XmNpageDecrementCallback);
  gh_new_procedure0_0 ("xm-n-ok-callback", gh_XmNokCallback);
  gh_new_procedure0_0 ("xm-n-cancel-callback", gh_XmNcancelCallback);
  gh_new_procedure0_0 ("xm-n-browse-selection-callback",
		       gh_XmNbrowseSelectionCallback);

  gh_new_procedure0_0 ("xm-drawing-area-widget-class",
		       gh_xmDrawingAreaWidgetClass);
  gh_new_procedure0_0 ("xm-form-widget-class", (ghp) gh_xmFormWidgetClass);
  gh_new_procedure0_0 ("xm-push-button-widget-class",
		       gh_xmPushButtonWidgetClass);
  gh_new_procedure0_0 ("xm-cascade-button-widget-class",
		       gh_xmCascadeButtonWidgetClass);
  gh_new_procedure0_0 ("xm-scroll-bar-widget-class",
		       gh_xmScrollBarWidgetClass);
  gh_new_procedure0_0 ("xm-file-selection-box-widget-class",
		       gh_xmFileSelectionBoxWidgetClass);
  gh_new_procedure0_0 ("xm-text-field-widget-class",
		       gh_xmTextFieldWidgetClass);

  gh_new_procedure0_0 ("xt-main-loop", (ghp) gh_XtMainLoop);
  gh_new_procedure0_1 ("dereference", (ghp) gh_dereference);
}
